/*
 * vifctl/iw_compat.c
 * $Rev: 1433 $
 *
 * Copyright (C) 2009 Reto Gantenbein <reto.gantenbein@gmail.com>
 * Copyright (C) 1997-2007 Jean Tourrilhes <jt@hpl.hp.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/**
 * @file iw_compat.c
 * @brief Compatibility functions for wireless structure handling
 *
 * This file contains functions for handling wireless ioctl() calls.
 * It is mainly derived from iwlib.c from the wireless-tools package.
 */

#include <net/ethernet.h>           // ether_addr
#include <math.h>                   // pow(), floor(), log10()
#include <stdio.h>                  // sprintf()
#include "iw_compat.h"


/**
 * Display an Ethernet address (MAC) in readable format.
 */
void
iw_ether_ntop(const struct ether_addr *eth, char *buf)
{
    sprintf(buf, "%02X:%02X:%02X:%02X:%02X:%02X",
        eth->ether_addr_octet[0], eth->ether_addr_octet[1],
        eth->ether_addr_octet[2], eth->ether_addr_octet[3],
        eth->ether_addr_octet[4], eth->ether_addr_octet[5]);
}

/**
 * Display an wireless access point socket address in readable format.
 * Note : 0x44 is an accident of history, that's what the Orinoco/PrismII
 * chipset report, and the driver doesn't filter it.
 */
char *
iw_sawap_ntop(const struct sockaddr *sap, char *buf)
{
    const struct ether_addr ether_zero = {{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }};
    const struct ether_addr ether_bcast = {{ 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF }};
    const struct ether_addr ether_hack = {{ 0x44, 0x44, 0x44, 0x44, 0x44, 0x44 }};
    const struct ether_addr *ether_wap = (const struct ether_addr *) sap->sa_data;

    if (!iw_ether_cmp(ether_wap, &ether_zero))
        sprintf(buf, "Not-Associated");
    else
        if (!iw_ether_cmp(ether_wap, &ether_bcast))
            sprintf(buf, "Invalid");
        else
            if(!iw_ether_cmp(ether_wap, &ether_hack))
                sprintf(buf, "None");
            else
                iw_ether_ntop(ether_wap, buf);

    return(buf);
}


/**
 * Convert a floating point the our internal representation of
 * frequencies.
 * The kernel doesn't want to hear about floating point, so we use
 * this custom format instead.
 */
void
iw_float2freq(double in, iwfreq *out)
{
    out->e = (short) (floor(log10(in)));
    if(out->e > 8)
    {
        out->m = ((long) (floor(in / pow(10,out->e - 6)))) * 100;
        out->e -= 8;
    } else {
      out->m = (long) in;
      out->e = 0;
    }
}

/**
 * Convert our internal representation of frequencies to a floating point.
 */
double
iw_freq2float(const iwfreq *in)
{
    return ((double) in->m) * pow(10,in->e);
}

/** Convert power values from dBm to mW. */
int16_t
iw_dbm2mw(int16_t in)
{
  return((int16_t) (floor(pow(10.0, (((double) in) / 10.0)))));
}

/** Convert a value in milliWatt to a value in dBm. */
int16_t
iw_mwatt2dbm(int16_t in)
{
  return((int) (ceil(10.0 * log10((double) in))));
}

/**
 * Return parameter state to set has_param value in struct wireless_info. This
 * function is not in the original iwlib.c.
 * @see struct wireless_info
 * @param par iw_param to retrieve state
 * @return 1 if parameter is enabled, 0 else
 */
int
iw_get_param_state(struct iw_param * par)
{
    return !(par->disabled == 1);
}