//
// Copyright (C) 2008-2010 Reto Gantenbein
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
//

#include "VifBackend.h"

Define_Module(VifBackend);

VifBackend::VifBackend()
{
	host = NULL;
	for (int i = 0; i < MAX_NUM_VIF; i++)
		vif[i] = NULL;
	memset(&addrInfo, 0, sizeof(struct addrinfo));
	memset(&hostAddr, 0, sizeof(struct sockaddr_storage));
	numVif = 0;
}

VifBackend::~VifBackend() {}

void VifBackend::initialize()
{
	host = check_and_cast<cModule *>(getParentModule());
	numVif = host->par("numVif");
	for (int i = 0; i < numVif; i++)
		vif[i] = check_and_cast<cModule *>(host->getSubmodule("ath", i));

	std::string hostIP = host->par("ipAddress");
	const char *ipp = hostIP.c_str();

	struct addrinfo request, *result ;
	memset(&request, 0, sizeof(struct addrinfo));
	request.ai_family = AF_UNSPEC;
	request.ai_socktype = SOCK_DGRAM;
	request.ai_flags = AI_NUMERICHOST | AI_NUMERICSERV;
	request.ai_protocol = 0;
	request.ai_canonname = NULL;
	request.ai_addr = NULL;
	request.ai_next = NULL;
	int port = host->par("port");
    char cport[5];
    sprintf(cport, "%i", port);
    if (getaddrinfo(ipp, cport, &request, &result) < 0)
    	throw new cRuntimeError("VifBackend: getaddrinfo");
    else {
    	memmove(&addrInfo, result, sizeof(struct addrinfo));
    	memmove(&hostAddr, result->ai_addr, sizeof(struct sockaddr_storage));
    	addrInfo.ai_addr = (struct sockaddr *) &hostAddr;
    }
}

void VifBackend::handleMessage(cMessage *etherFrame)
{
	ev << getFullPath() << ".handleMessage(): msg received!\n";   //TODO: for debug only!

	// we should only receive messages from lower layer
	RAWEtherFrame *frame = check_and_cast<RAWEtherFrame *>(etherFrame);

    int payloadSize = frame->getRawBytesArraySize();
    int messageSize = DATA_PREAMBLE_SIZE + payloadSize;

    // create wlan_msg
    dataMsg *msg = (dataMsg *) calloc(1, messageSize);
    ((msgHeader *) msg)->type = DATA_MSG;
    strcpy(((msgHeader *) msg)->id, "0\0");				//TODO: WlanModel id?
    char *payloadPointer = (char *) &(msg->wlanPkt);
    frame->getPayload(payloadPointer);
    msg->pktLength = payloadSize;

    // send back to host
    sendToHost((char *) msg, messageSize);

    // log packet processing time
	timeval curTime;
	gettimeofday(&curTime, NULL);
	timeval processingTimeval = timeval_substract(curTime, (frame->getTimestamp()).dbl());
	double processingTime = (processingTimeval.tv_sec + processingTimeval.tv_usec * 1e-6) * 1000;  // in ms
	ev << getFullPath() << ".handleReply(): processingTime " << processingTime << "ms\n";
	//pktProcTime.record(processingTime);				//TODO: add logging

    delete frame;
    free(msg);
}

void VifBackend::sendAck(unsigned int msgId)
{
	Enter_Method("sendAck()");
    plainMsg *ack;

    ack = (plainMsg *) calloc(1, ACK_MSG_SIZE);
    ack->msgId = msgId;
    ((msgHeader *) ack)->type = ACK_MSG;

//   ProtocolHandler::printPlainMessage((char *) ack);			//TODO: only for debug!
    ev << getFullPath() << ": send ACK_MSG (id: " << msgId << ") to host\n";

    sendToHost((char *) ack, ACK_MSG_SIZE);

    free(ack);
}

void VifBackend::sendToHost(char *msg, int length)
{
	EmulationRTScheduler *rtScheduler =
		check_and_cast<EmulationRTScheduler *>(simulation.getScheduler());
	rtScheduler->sendMessage(msg, length, (struct sockaddr *) &hostAddr,
							 addrInfo.ai_addrlen);
}

void VifBackend::sendToModel(char *wlanPkt, int length)
{
	Enter_Method("sendToModel()");
    RAWEtherFrame *etherFrame = new RAWEtherFrame("etherFrame");

//    etherFrame->setTimestamp(msg->getTimestamp());

    // fill RAWEtherFrame message
    etherFrame->setByteLength(length);
    etherFrame->setRawBytesArraySize(length);
    etherFrame->setPayload(wlanPkt);
    etherFrame->setDest();

    // print some information
    MACAddress destAddr = (((Ieee802Ctrl *) etherFrame->getControlInfo())->getDest());
    ev << host->getName() << ":VifBackend.sendToModel(): sending EtherFrame to " << destAddr.str() << "\n";

    // send via wireless stack (see VirtualHost.ned for details)
    send(etherFrame, "ifOut");
}

void VifBackend::setRTS(int8_t index, int16_t rts)
{
	ev << getFullPath() << ": set RTS/CTS threshold to " << rts << " Bytes\n";
	cModule *mac = check_and_cast<cModule *>(vif[index]->getSubmodule("mac"));
	mac->par("rtsThresholdBytes") = rts;
}

// The bitrate and the receive/transmit channel can be changed by sending Radio
// a blank cMessage with PHY_C_CONFIGURERADIO as message kind, and a PhyControlInfo
// object attached to it as control info, to hold the new bitrate and/or channel number.
// If the radio is currently transmitting, the change will actually take place after the
// transmission is completed.
void VifBackend::setChannel(int8_t index, int16_t channel)
{
	ev << getFullPath() << ": set channel to " << channel << "\n";
	PhyControlInfo *phyCtrl = new PhyControlInfo();
	phyCtrl->setChannelNumber(channel);
	cMessage *msg = new cMessage("changeChannel", PHY_C_CONFIGURERADIO);
	msg->setControlInfo(phyCtrl);
	send(msg, "ifOut");
}

void VifBackend::setRetry(int8_t index, int16_t retry)
{
	ev << getFullPath() << ": set number of retries to " << retry << "\n";
	cModule *mac = check_and_cast<cModule *>(vif[index]->getSubmodule("mac"));
	mac->par("retryLimit") = retry;
}

void VifBackend::setTxPower(int8_t index, int16_t txpower)
{
	ev << getFullPath() << ": set transmission power to " << txpower << " mW\n";
	cModule *radio = check_and_cast<cModule *>(vif[index]->getSubmodule("radio"));
	radio->par("transmitterPower") = txpower;
}

void VifBackend::setSensitivity(int8_t index, int16_t sensitivity)
{
    // Since iwlib uses dBm to express sensitivity we cannot have very detailed values
	// in mW., E.g. 85mW -> mW2dBm -> 20dBm -> dBm2mW -> 100mW
	// So we use the proprietary positive sensitivity code to express fine grade mW values
	if (sensitivity == 1)
		sensitivity = -85;
	ev << getFullPath() << ": set sensitivity value to " << sensitivity << " mW\n";
	cModule *radio = check_and_cast<cModule *>(vif[index]->getSubmodule("radio"));
	radio->par("sensitivity") = sensitivity;
}

void VifBackend::updateVifParameter(configNotification *info)
{
	Enter_Method("updateVifParameter()");

	int vifIndex = 0; 	//TODO: set correct index
	if (vifIndex > numVif)
	{
		ev << getFullPath() << ": No interface with index " << vifIndex << "!\n";
		return;
	}
	switch(info->type)
	{
		// INFO: set fragmentation is not supported by model
		case SIOCSIWFREQ:
			setChannel(vifIndex, info->value.channel);
			break;
		case SIOCSIWRTS:
			setRTS(vifIndex, info->value.rts);
			break;
		case SIOCSIWRETRY:
			setRetry(vifIndex, info->value.retry);
			break;
		case SIOCSIWTXPOW:
			setTxPower(vifIndex, info->value.txpower);
			break;
		case SIOCSIWSENS:
			setSensitivity(vifIndex, info->value.sensitivity);
			break;
		default:
			ev << "VifBackend.updateVifParameter(): Unknown ioctl() " << info->type << "\n";
			break;
	}
}

void VifBackend::initializeVifParameters(vifRegistrationData vifReg[])
{
	Enter_Method("initializeVifParameters()");

	for (int i = 0; i < numVif; i++)
	{
		// INFO: set fragmentation is not supported by model
		setRTS(i, vifReg->rts);
		setRetry(i, vifReg->retry);
		setTxPower(i, vifReg->txpower);
		setChannel(i, vifReg->channel);
		setSensitivity(i, vifReg->sensitivity);
	}
}
